# General packages
library(dplyr)
library(tidyr)
library(furrr)

# Geo packages
library(terra)
library(CoordinateCleaner)
library(sf)
library(geos)

# Stat packages
library(MASS)
library(dismo)
library(spatialEco)
library(blockCV)
library(caret)

source("R/utils.R")

load("data/r_objects/range_maps.RData")
load("data/r_objects/occs_final.RData")

# ---------------------------------------------------------------------------#
# Prepare Geodata                                                         ####
# ---------------------------------------------------------------------------#
target_species = unique(occs_final$species)

sa_polygon = rnaturalearth::ne_countries() %>% 
  dplyr::filter(continent == "South America") %>% 
  sf::st_union() 

raster_filepaths = list.files("~/symobio-modeling/data/geospatial/raster/", full.names = T) %>% 
  #raster_filepaths = list.files("~/share/groups/mas_data/00_data/processed/CHELSA_v2-1_bioclim/", pattern = ".tif$", full.names = T) %>% # TODO: finalize predictor choice
  stringr::str_sort(numeric = T) 

# Project (proj string generated with SA bbox coordinates on https://projectionwizard.org)
proj_string = "+proj=tcea +lon_0=-58.0704167 +datum=WGS84 +units=m +no_defs" 

# raster_data = terra::project(raster_data, proj_string)
sa_polygon = st_transform(sa_polygon, proj_string)
occs_final = st_transform(occs_final, proj_string)
range_maps = st_transform(range_maps, proj_string)

# ---------------------------------------------------------------------------#
# Sample absences                                                         ####
#                                                                            #
# Absences should be predominantly sampled from regions that are             #
# geographically close (reproduce sampling biases) but environmentally       #
# dissimilar (avoid false negatives) to the known occurrences                #
# ---------------------------------------------------------------------------#
future::plan("multisession", workers = 16)

model_data = furrr::future_walk(
  .x = target_species,
  .options = furrr::furrr_options(seed = 42, scheduling = FALSE), # make sure workers get constant stream of work
  .env_globals = c(raster_filepaths, sa_polygon, occs_final, range_maps, proj_string),
  .f = function(species){
    # Prepare species occurrence data #####
    occs_spec = dplyr::filter(occs_final, species == !!species)
    if(nrow(occs_spec) <= 1){
      warning("Skipping species ", species, ": Can't process species with a single presence record." )
      return()
    }
    
    occs_multipoint = occs_spec %>% 
      summarise(geometry = st_combine(geometry))
    
    range_spec = dplyr::filter(range_maps, name_matched == !!species) %>% 
      rename(species = name_matched)
    
    # Define model/sampling region
    # 1. Union all points from range polygon and occurrences records
    # 2. Build concave polygon from unionen points
    # 3. Buffer by 100 km to extend sampling region into new environments
    # 4. Crop by South America outline
    sampling_region = st_cast(range_spec, "MULTIPOINT") %>% 
      st_union(occs_multipoint) %>% 
      geos::geos_concave_hull(ratio = 0.3) %>% # Concave hull in sf requires geos 3.11
      st_as_sf() %>% 
      st_set_crs(proj_string) %>% 
      st_buffer(dist = 100000) %>% 
      st_intersection(sa_polygon) 
    
    # Load raster data 
    raster_data = terra::rast(raster_filepaths) %>% 
      terra::project(proj_string) %>% 
      terra::crop(sf::st_bbox(sampling_region))
    
    # Define spatial KDE
    ref = st_bbox(sampling_region)[c(1,3,2,4)]
    
    # Function to estimate bandwidth 
    # ---> this function is the default estimation of the sf.kde package but with a minimum of 50000 = 50km)
    calc_bw <- function(x) {
      r <- stats::quantile(x, c(0.25, 0.75))
      h <- (r[2] - r[1])/1.34
      result = 4 * 1.06 * min(sqrt(stats::var(x)), h) * length(x)^(-1/5)
      return(max(result, 50000))
    }
    
    bw_x = calc_bw(st_coordinates(occs_spec)[,"X"])
    bw_y = calc_bw(st_coordinates(occs_spec)[,"Y"])
    spatial_kde = spatialEco::sf.kde(occs_spec, bw = c(bw_x, bw_y), res = 10000, ref = ref, standardize = T, scale.factor = 1) %>% 
      crop(sampling_region, mask = T, touches = F)
    
    # Define reverse niche
    env_sampling_region = terra::crop(raster_data, sampling_region, mask = T) %>% 
      terra::resample(spatial_kde) %>% 
      raster::stack()
    
    maxent_fit <- maxent(env_sampling_region, st_coordinates(occs_spec))
    env_kde <- (1-predict(env_sampling_region, maxent_fit)) %>% 
      terra::rast()
    
    # Sample pseudo absences
    abs_prob = spatial_kde * env_kde
    abs_spec_list = list()
    
    samples_required = nrow(occs_spec)
    
    while(samples_required != 0){
      sample_points = spatSample(abs_prob, size = samples_required, method = "weights", xy = T, as.df = T) %>% 
        dplyr::select(-z) %>% 
        dplyr::mutate(x = x + runif(nrow(.), -5000, 5000),         # Add jitter (res/2) to cell centroids 
                      y = y + runif(nrow(.), -5000, 5000)) %>%     # Add jitter (res/2) to cell centroids 
        st_as_sf(coords = c("x", "y"), crs = proj_string)
      
      sample_abs = bind_cols(
        terra::extract(raster_data, terra::vect(sample_points), ID = FALSE)
      ) %>% 
        bind_cols(sample_points) %>% 
        drop_na()
      
      abs_spec_list[[length(abs_spec_list)+1]] = sample_points
      
      samples_required = samples_required - nrow(sample_points) # Sometimes there are no env data for sample points, so keep sampling
    }
    
    abs_spec = bind_rows(abs_spec_list)
    abs_spec = abs_spec %>% 
      bind_cols(
        st_coordinates(st_transform(abs_spec, "EPSG:4326")) 
      ) %>% 
      dplyr::rename(
        longitude = X,
        latitude = Y
      ) %>% 
      dplyr::mutate(
        species = !!species,
        present = 0
      ) 
    
    # Create presence-absence dataframe
    pa_spec = occs_spec %>% 
      dplyr::mutate(present = 1) %>% 
      bind_rows(abs_spec) 
    
    ggplot() +
      ggtitle(species) +
      geom_sf(data = st_crop(sa_polygon, ref), fill = "#756445") +
      tidyterra::geom_spatraster(data = abs_prob) +
      scale_fill_gradient(low="darkblue", high="green", na.value=NA, name = "Absence sampling probability") +
      geom_sf(data = range_spec, alpha = 0, color = "black") +
      geom_sf(data = pa_spec, aes(color = as.factor(present)), size = 0.7) +
      scale_color_discrete(name = "Presence") +
      theme_minimal()
    
    try({
      ggsave(paste0(species, ".pdf"), path = "plots/pa_sampling", device = "pdf", scale = 2)
    })
    
    # Define cross-validation folds
    folds = tryCatch({
      spatial_folds = suppressMessages(
        blockCV::cv_spatial(
          pa_spec,
          column = "present",
          k = 5,
          progress = F, plot = F, report = F
        )
      )
      
      spatial_folds$folds_ids
    }, warning = function(w){
      NA
    }, error = function(e){
      NA
    })
    
    pa_spec$fold_eval = folds
    pa_spec$geometry = NULL
    
    save(pa_spec, file = paste0("data/r_objects/pa_sampling/", species, ".RData"))
  }
)

# Combine presence-absence data
files = list.files("data/r_objects/pa_sampling/", full.names = T)
model_data = lapply(files, function(f){load(f); return(pa_spec)}) %>% 
  bind_rows()

save(model_data, file = "data/r_objects/model_data_pa_sampling.RData")
