library(dplyr)
library(tidyr)
library(cito)

source("R/utils.R")

load("data/r_objects/model_data.RData")
load("data/r_objects/func_dist.RData")
load("data/r_objects/phylo_dist.RData")
load("data/r_objects/range_dist.RData")

# ----------------------------------------------------------------------#
# Prepare data                                                       ####
# ----------------------------------------------------------------------#
model_species = Reduce(
  intersect, 
  list(unique(model_data$species), colnames(range_dist), colnames(phylo_dist), colnames(func_dist))
) 

model_data_final = model_data %>%
  dplyr::filter(species %in% !!model_species) %>% 
  dplyr::mutate(species_int = as.integer(as.factor(species)))

train_data = dplyr::filter(model_data_final, train == 1)
test_data = dplyr::filter(model_data_final, train == 0)

# Create embeddings
func_ind = match(model_species, colnames(func_dist))
func_dist = func_dist[func_ind, func_ind]
func_embeddings = eigen(func_dist)$vectors[,1:20]

phylo_ind = match(model_species, colnames(phylo_dist))
phylo_dist = phylo_dist[phylo_ind, phylo_ind]
phylo_embeddings = eigen(phylo_dist)$vectors[,1:20]

range_ind = match(model_species, colnames(range_dist))
range_dist = range_dist[range_ind, range_ind]
range_embeddings = eigen(range_dist)$vectors[,1:20]


# ----------------------------------------------------------------------#
# Train model                                                        ####
# ----------------------------------------------------------------------#
predictors = paste0("layer_", 1:19)

formula = as.formula(
  paste0("present ~ ", 
         paste(predictors, collapse = '+'),  
         " + e(species_int, weights = func_embeddings, lambda = 0.00001, train = F)",
         " + e(species_int, weights = phylo_embeddings, lambda = 0.00001, train = F)",
         " + e(species_int, weights = range_embeddings, lambda = 0.00001, train = F)"
  )
)

plot(1, type="n", xlab="", ylab="", xlim=c(0, 25000), ylim=c(0, 0.7)) # empty plot with better limits, draw points in there
msdm_fit_embedding_multi_nolonlat = dnn(
  formula,
  data = train_data,
  hidden = c(500L, 500L, 500L),
  loss = "binomial",
  activation = c("sigmoid", "leaky_relu", "leaky_relu"),
  epochs = 30000L, 
  lr = 0.01,   
  baseloss = 1,
  batchsize = nrow(train_data),
  dropout = 0.1,
  burnin = 100,
  optimizer = config_optimizer("adam", weight_decay = 0.001),
  lr_scheduler = config_lr_scheduler("reduce_on_plateau", patience = 100, factor = 0.7),
  early_stopping = 250,
  validation = 0.3,
  device = "cuda",
)
save(msdm_fit_embedding_multi_nolonlat, file = "data/r_objects/msdm_fit_embedding_multi_nolonlat.RData")

# ----------------------------------------------------------------------#
# Evaluate results                                                   ####
# ----------------------------------------------------------------------#
load("data/r_objects/msdm_fit_embedding_multi_nolonlat.RData")
data_split = test_data %>% 
  group_by(species_int) %>% 
  group_split()

msdm_results_embedding_multi_nolonlat = lapply(data_split, function(data_spec){
  target_species =  data_spec$species[1]
  data_spec = dplyr::select(data_spec, -species)
  
  msdm_performance = tryCatch({
    evaluate_model(msdm_fit_embedding_multi_nolonlat, data_spec)
  }, error = function(e){
    list(AUC = NA, Accuracy = NA, Kappa = NA, Precision = NA, Recall = NA, F1 = NA)
  })
  
  performance_summary = tibble(
    species = !!target_species,
    obs = length(which(model_data$species == target_species)),
    model = "MSDM_embed_informed_multi_nolonlat",
    auc = msdm_performance$AUC,
    accuracy = msdm_performance$Accuracy,
    kappa = msdm_performance$Kappa,
    precision = msdm_performance$Precision,
    recall = msdm_performance$Recall,
    f1 = msdm_performance$F1
  )
}) %>% bind_rows()

save(msdm_results_embedding_multi_nolonlat, file = "data/r_objects/msdm_results_embedding_multi_nolonlat.RData")